' Projekteinstellungen:
' Option Explicit On
' Option Strict On
' Imports Microsoft.VisualBasic.ControlChars
' Imports System
Imports System.Windows.Forms

Public Module modMain

    Public Sub Main()

        ' Array-Deklaration - beachte, da der oberste gltige Index (UpperBound) um eines
        ' geringer ist als die Anzahl (Length) der Elemente. 
        ' Dieses, weil auch die 0 ein gltiger Index ist
        Const Length0 As Integer = 10
        Dim Arr0(Length0 - 1) As Integer
        ' zwei Varianten, die Elemente zu enumerieren
        For I As Integer = 0 To Length0 - 1
            Arr0(I) = I
        Next
        For Each Value As Integer In Arr0
            WriteLine(Value)
        Next
        WriteLine()

        ' Bei der initialisierenden Deklaration entfllt die Angabe der Anzahl der Elemente - das
        ' "sieht" der Compiler selbst
        Dim Words As String() = New String() {"If", "Then", "Else"}
        ' Man kann auch knapper formulieren
        Dim SmartWords As String() = {"If", "Then", "Else"}

        ' noch einmal der Unterschied zwischen UpperBound und Lnge eines Arrays:
        WriteLine("UpperBound: ", Words.GetUpperBound(0), ", Length: ", Words.Length, Lf)

        ' Arrays haben die schnellsten Zugriffszeiten, den geringsten Speicherverbrauch, und die
        ' Elemente sind typisiert.
        ' Auerdem bietet die Array-Klasse einen umfangreichen Satz schneller und ntzlicher
        ' Methoden, (auf die ich hier nicht eingehe): IndexOf(), Find(), BinarySearch(), Sort(),
        ' Reverse(), CopyTo().

        ' ## Es knnen aber keine Elemente hinzugefgt oder entfernt werden. ##
        ' Redimensionieren erstellt ein neues Array, und kopiert ggfs. (Preserve) die Werte um.
        ' Das ist manchmal im Einzelfall sinnvoll (Optimierung auf Zugriffszeiten), aber v.a.
        ' Schleifen wie diese sind auerordentlich unperformant:
        For I As Integer = 0 To Length0 - 1
            ReDim Preserve Arr0(I)
            Arr0(I) = I
        Next
        ' stattdessen kann man die List(Of T) verwenden:
        Dim List0 As New List(Of Integer)
        For I As Integer = 0 To Length0 - 1
            List0.Add(I)
        Next
        ' Auf List(Of T) kann man dieselben beiden For-Schleifen anwenden wie auf T-Array

        ' List(Of T) kann man auch initialisieren - etwa mit einem Array
        Dim List1 As New List(Of String)(New String() {"If", "Then", "Else"})

        ' Leider kann man List(Of T) nicht mit einem ParamArray initialisieren
        ' Da schafft ein kleines Helfer-Funktinchen etwas Abhilfe
        Dim List2 As List(Of String) = NewList("If", "Then", "Else")

        ' Kommen wir zur Mehrdimensionalitt
        Call MultiDimensions()
    End Sub

    Public Function NewList(Of T)(ByVal ParamArray Items() As T) As List(Of T)
        Return New List(Of T)(Items)
    End Function

    Sub MultiDimensions()
        Dim Arr0(3, 1) As Integer
        ' Bei "echter Mehrdimensionalitt" im Array reicht die Array.Length-Property nicht mehr
        ' aus, man mu den UpperBound der einzelnen Dimensionen ermitteln
        Dim YUbound As Integer = Arr0.GetUpperBound(0)
        Dim XUbound As Integer = Arr0.GetUpperBound(1)
        For Y As Integer = 0 To YUbound
            For X As Integer = 0 To XUbound
                '(zur Erinnerung: UpperBound ist eins kleiner als die Lnge des Arrays)
                Arr0(Y, X) = Y * (XUbound + 1) + X
            Next
        Next
        ' Kuriose ForEach-Iteration: Arr0 wird wie ein eindimensionales Array ausgelesen
        For Each Numb As Integer In Arr0
            WriteLine(Numb)
        Next

        ' Merkspruch: "Zeile zuerst, Spalte spter"
        ' Es geht darum, da es mathematisch gesehen egal ist, ob ein 2-dimensionales Array seine
        ' Elemente zunchst von oben nach unten durchgeht, und sich dann die nchste Spalte
        ' vornimmt, oder ob es von links nach rechts liest, und dann die nchste Zeile. 
        ' Zu vermeiden ist nur, sein Array erst spaltenweise aufzubauen, und spter dann
        ' zeilenweise auszulesen (auer, man beabsichtigt dieses explizit).
        ' Daher o.g. Konvention, der auch der Bildschirm-Aufbau folgt, oder etwa die alltgliche
        ' Lese-Richtung in der westlichen Kultur.
        ' Und bei derlei Arrays bedeuten die Indizees eben: "Zeile zuerst, Spalte spter"

        ' Beweis: Initialisiere String(2, 1) ( = 3 Zeilen, 2 Spalten )
        Dim ArrDic As String(,) = {{"If", "Wenn"}, {"Then", "Dann"}, {"Else", "Andernfalls"}}
        ' Der Merkspruch gilt eben auch fr den Aufbau verschachtelter Schleifen.
        For Y As Integer = 0 To 2           ' Zeile zuerst
            For X As Integer = 0 To 1       ' Spalte spter
                Write(ArrDic(Y, X))
            Next
            WriteLine()
        Next
        WriteLine()
        ' Fallstricke lauern beim Umgang mit Grafik und mit Tabellen-Controls, da diese gegen o.g.
        ' Merkspruch verstoen:
        Dim Pt As New System.Drawing.Point(X:=3, Y:=2)
        Dim DGV As New DataGridView : DGV.RowCount = 1 : DGV.ColumnCount = 1
        Dim C As DataGridViewCell = DGV(columnIndex:=0, rowIndex:=0)

        ' zur Vertiefung die Initialisierung eines 3-dimensionalen Arrays - beachten Sie den
        ' Einsatz der geschweiften Klammern
        Dim Wuerfel(,,) As Integer = {{{0, 1}, {2, 3}}, {{4, 5}, {6, 7}}}
        '( Upperbound = 1, in allen 3 Dimensionen )
        For Z As Integer = 0 To 1
            For Y As Integer = 0 To 1
                For X As Integer = 0 To 1
                    Write(Wuerfel(Z, Y, X))
                Next
                WriteLine()
            Next
            WriteLine()
        Next
        WriteLine()

        ' "unechte Mehrdimensionalitt" - Array von Arrays
        ' Angenommen, Sie brauchen ein Array, da in Zeile Eins drei Eintrge enthlt, in Zeile
        ' zwei hingegen vier?
        ' Verwenden Sie ein Array von Arrays:
        Dim KeyWords As String()() = New String()() { _
            New String() {"If", "Else", "End If"}, _
            New String() {"Try", "Catch", "Finally", "End Try"}, _
            New String() {"Do", "Loop"}, _
            New String() {"For", "Next"}, _
            New String() {"Imports"}}

        For Y As Integer = 0 To KeyWords.Length - 1
            For X As Integer = 0 To KeyWords(Y).Length - 1
                ' Beim Zugriff hat jeder Index seine eigene Klammer, im Gegensatz zur "echten"
                ' Mehrdimensionalitt
                Write(KeyWords(Y)(X))
            Next
            WriteLine()
        Next
        WriteLine()

        ' ForEach mal wieder eleganter:
        For Each Line As String() In KeyWords
            For Each Word As String In Line
                Write(Word)
            Next
            WriteLine()
        Next
        WriteLine()

        '"unechte" Mehrdimensionalitt mit List(Of T):
        Dim KeyWordList As List(Of List(Of String)) = NewList( _
             NewList("If", "Else", "End If"), _
             NewList("Try", "Catch", "Finally", "End Try"), _
             NewList("Do", "Loop"), _
             NewList("For", "Next"), _
             NewList("Imports"))

        'spaeshalber mal ohne NewList()-Funktion (alle Listen mit Arrays initialisieren):
        KeyWordList = New List(Of List(Of String))(New List(Of String)() { _
            New List(Of String)(New String() {"If", "Else", "End If"}), _
            New List(Of String)(New String() {"Try", "Catch", "Finally", "End Try"}), _
            New List(Of String)(New String() {"Do", "Loop"}), _
            New List(Of String)(New String() {"For", "Next"}), _
            New List(Of String)(New String() {"Imports"})})
        ' So eine Initialisierung ist nun wirklich nicht mehr sehr leserlich

        ' Dafr kann man aber sowohl Worte als auch ganze Zeilen nachtrglich einfgen oder
        ' entfernen:
        KeyWordList(0).Insert(1, "Then")                   ' einer Gruppe ein Element zufgen
        KeyWordList.Insert(2, NewList("While", "Wend"))    ' eine Gruppe zufgen
        KeyWordList.RemoveAt(KeyWordList.Count - 1)        ' letzte Gruppe entfernen

        ' Der Zugriff ist identisch mit dem Zugriff bei Array-Arrays
        For Y As Integer = 0 To KeyWordList.Count - 1
            For X As Integer = 0 To KeyWordList(Y).Count - 1
                Write(KeyWordList(Y)(X))
            Next
            WriteLine()
        Next
        WriteLine()

        Call Dictionaries()
    End Sub


    Sub Dictionaries()
        ' Whrend Arrays und Listen einen numerisch indizierten Zugriff gewhren, bietet das
        ' Dictionary einen Zugriff ber Schlssel-**Objekte**
        ' Den Datentyp von Schlssel und Wert kann man per generischem Parameter festlegen; hier
        ' ein einfaches Beispiel fr Strings:
        Dim Dic As New Dictionary(Of String, String)
        With Dic
            .Add("If", "Wenn")
            .Add("Then", "Dann")
            .Add("Else", "Andernfalls")
        End With

        WriteLine("Dic(""If"") = ", Dic("If"))

        ' Eine manchmal(!) praktische Kuriositt beim Schreibzugriff: Existiert der Schlssel noch
        ' gar nicht, so wird er immanent angelegt.
        Dic("End If") = "Ende Wenn"

        ' Beim Lese-Zugriff auf einen nicht vorhandenen Schlssel erfolgt natrlich eine
        ' KeyNotFoundException 
        Dim Response As String = Dic("Try")

        ' Dictionaries sind ebenfalls enorm schnell im Zugriff.
        ' Intern werden die Werte nach dem sog. Hashing-Verfahren "eingerumt", was einen
        ' direkten Abruf ohne Such-Algorithmus ermglicht

        ' Enumerieren kann man ein Dictionary auch, sogar in mehrfacher Hinsicht:
        For Each Entry As KeyValuePair(Of String, String) In Dic
            WriteLine(Entry.Key, " = ", Entry.Value)
        Next
        ' oder
        For Each Key As String In Dic.Keys
            WriteLine(Key)
        Next
        ' oder
        For Each Value As String In Dic.Values
            WriteLine(Value)
        Next
        ' anders als bei Listen haben wir hier aber **keinen numerischen Index**
        ' Dim P As KeyValuePair(Of String, String) = Dic(2) '<- geht nicht

        ' Beachten Sie, da die .Keys() und .Values()-Property jeweils ein **Export** des
        ' Dictionaries ist. 
        ' D.h., bei jedem Abruf wird intern eine Collection erstellt und befllt, also das gesamte
        ' Dictionary durchlaufen.
        ' folgender Code ist also 3 mal performanter (an Geschwindigkeit *und* Speicherplatz) als
        ' der darauf folgende:
        Dim Values As Dictionary(Of String, String).ValueCollection = Dic.Values
        For I As Integer = 0 To 2
            For Each Val As String In Values
                ' machwas
            Next
        Next

        ' unperformant
        For I As Integer = 0 To 2
            For Each Val As String In Dic.Values
                ' machwas
            Next
        Next

    End Sub

End Module

''' <summary>zwei kleine Helferlein zur vereinfachten Ausgabe</summary>
Public Module modHelpers
    Public Sub WriteLine(ByVal ParamArray Args() As Object)
        Console.WriteLine(String.Concat(Args))
    End Sub
    Public Sub Write(ByVal Arg As Object)
        Console.Write(String.Concat(Arg, " "))
    End Sub
End Module
